/*
hts.c

Copyright (C) 1999 Lars Brinkhoff.  See COPYING for terms and conditions.

hts is the server half of httptunnel.  httptunnel creates a virtual
two-way data path tunneled in HTTP requests.
*/

#include <stdio.h>
#include <fcntl.h>
#include <stdlib.h>
#include <sys/poll.h>

#include "common.h"

typedef struct
{
  char *me;
  char *device;
  int port;
  char *forward_host;
  int forward_port;
} Arguments;

int debug_level = 0;
FILE *debug_file = NULL;

static void
usage (FILE *f, const char *me)
{
  fprintf (f,
"Usage: %s [OPTION]... [PORT]\n"
"Listen for incoming httptunnel connections at PORT (default port is %d).\n"
"When a connection is made, I/O is redirected to the destination specified\n"
"by the --device or --forward-port switch.\n"
"\n"
"  -d, --device DEVICE            use DEVICE for input and output\n"
#ifdef DEBUG
"  -D, --debug [LEVEL]            enable debug mode\n"
#endif
"  -f, --forward-port HOST:PORT   connect to PORT at HOST and use it for \n"
"                                 input and output\n"
"  -h, --help                     display this help and exit\n"
#ifdef DEBUG
"  -l, --logfile FILE             specify logfile for debug output\n"
#endif
"  -V, --version                  output version information and exit\n"
"\n"
"Report bugs to %s.\n",
	   me, DEFAULT_HOST_PORT, BUG_REPORT_EMAIL);
}

static void
parse_arguments (int argc, char **argv, Arguments *arg)
{
  int c;

  /* defaults */

  arg->me = argv[0];
  arg->port = DEFAULT_HOST_PORT;
  arg->device = NULL;
  arg->forward_host = NULL;
  arg->forward_port = -1;
  
  for (;;)
    {
#ifdef HAVE_GETOPT_LONG
      int option_index = 0;
      static struct option long_options[] =
      {
	{ "help", no_argument, 0, 'h' },
	{ "version", no_argument, 0, 'V' },
#ifdef DEBUG
	{ "debug", optional_argument, 0, 'D' },
	{ "logfile", required_argument, 0, 'l' },
#endif
	{ "device", required_argument, 0, 'd' },
	{ "forward-port", required_argument, 0, 'f' },
	{ 0, 0, 0, 0 }
      };
#endif /* HAVE_GETOPT_LONG */

#ifdef DEBUG
      static const char *short_options = "d:D:f:hl:V";
#else
      static const char *short_options = "d:f:hV";
#endif

      c = getopt_long (argc, argv, short_options,
		       long_options, &option_index);
      if (c == -1)
	break;

      switch (c)
	{
#ifdef HAVE_GETOPT_LONG
	case 0:
	  printf ("option %s", long_options[option_index].name);
	  if (optarg)
	    printf (" with arg %s", optarg);
	  printf ("\n");
	  break;
#endif /* HAVE_GETOPT_LONG */
	  
	case 'd':
	  arg->device = optarg;
	  break;

#ifdef DEBUG
	case 'D':
	  if (optarg)
	    debug_level = atoi (optarg);
	  else
	    debug_level = 1;
	  break;

	case 'l':
	  debug_file = fopen (optarg, "w");
	  if (debug_file == NULL)
	    {
	      fprintf (stderr, "%s: couldn't open file %s for writing\n",
		       arg->me, optarg);
	      exit (1);
	    }
	  break;
#endif /* DEBUG */

	case 'f':
	  name_and_port (optarg, &arg->forward_host, &arg->forward_port);
	  if (arg->forward_port == -1)
	    {
	      fprintf (stderr, "%s: you must specify a port number.\n"
		               "%s: try '%s --help' for help.\n",
		       arg->me, arg->me, arg->me);
	      exit (1);
	    }
	  break;

	case 'h':
	  usage (stdout, arg->me);
	  exit (0);

	case 'V':
	  printf ("hts (%s) %s\n", PACKAGE, VERSION);
	  exit (0);

	case '?':
	  break;

	default:
	  printf ("?? getopt returned character code 0%o ??\n", c);
	}
    }

  if (argc - 1 == optind)
    arg->port = atoi (argv[optind]);
  else if (argc - 1 > optind)
    {
      usage (stderr, arg->me);
      exit (1);
    }

  if (arg->device == NULL && arg->forward_port == -1)
    {
      fprintf (stderr, "%s: one of --device or --forward-port must be used.\n"
	               "%s: try '%s -help' for help.\n",
	       arg->me, arg->me, arg->me);
      exit (1);
    }

  if (arg->device != NULL && arg->forward_port != -1)
    {
      fprintf (stderr, "%s: --device can't be used together with "
	                   "--forward-port.\n"
	               "%s: try '%s --help' for help.\n",
	       arg->me, arg->me, arg->me);
      exit (1);
    }

  if (debug_level == 0 && debug_file != NULL)
    {
      fprintf (stderr, "%s: --logfile can't be used without --debug\n",
	       arg->me);
      exit (1);
    }

  if (((arg->device == NULL) == (arg->forward_port == -1)) ||
      arg->port == -1 ||
      ((arg->forward_host == NULL) != (arg->forward_port == -1)))
    {
      usage (stderr, arg->me);
      exit (1);
    }
}

int
main (int argc, char **argv)
{
  int closed;
  int fd = -1;
  Arguments arg;
  Tunnel *tunnel;

  parse_arguments (argc, argv, &arg);

  if (debug_level == 0 || debug_file != NULL)
    daemon (0, 1);

#ifdef DEBUG
  if (debug_level > 0 && debug_file == NULL)
    debug_file = stdout;
#else
  openlog ("hts", LOG_PID, LOG_DAEMON);
#endif

  notice ("started with arguments:");
  notice ("  me = %s", arg.me);
  notice ("  device = %s", arg.device);
  notice ("  port = %d", arg.port);
  notice ("  forward_port = %d", arg.forward_port);
  notice ("  forward_host = %s", arg.forward_host);

  tunnel = tunnel_new_server (arg.port);
  if (tunnel == NULL)
    {
      fprintf (stderr, "%s: couldn't create tunnel\n", argv[0]);
      exit (1);
    }

  for (;;)
    {
      debug ("waiting for tunnel connection\n");

      if (arg.device != NULL)
	{
	  fd = open_device (arg.device);
	  debug ("open_device (\"%s\") = %d\n", arg.device, fd);
	  if (fd == -1)
	    {
	      fprintf (stderr, "%s: couldn't open %s: %s\n",
		       arg.me, arg.device, strerror (errno));
	      exit (1);
	    }
	}

      if (tunnel_accept (tunnel) == -1)
	{
	  notice ("couldn't accept connection: %s", strerror (errno));
	  continue;
	}
      notice ("accepted tunnel connection");

      if (arg.forward_port != -1)
	{
	  struct sockaddr_in addr;

	  if (set_address (&addr, arg.forward_host, arg.forward_port) == -1)
	    {
	      fprintf (stderr, "%s: couldn't forward port to %s:%d: %s\n",
		       arg.me, arg.forward_host, arg.forward_port,
		       strerror (errno));
	      exit (1);
	    }

	  fd = do_connect (&addr);
	  debug ("do_connect (\"%s:%d\") = %d\n",
		 arg.forward_host, arg.forward_port, fd);
	  if (fd == -1)
	    {
	      fprintf (stderr, "%s: couldn't connect to %s:%d: %s\n",
		       arg.me, arg.forward_host, arg.forward_port,
		       strerror (errno));
	      exit (1);
	    }
	}

      closed = FALSE;
      while (!closed)
	{
	  struct pollfd pollfd[2];
	  int n;
      
	  pollfd[0].fd = fd;
	  pollfd[0].events = POLLIN;
	  pollfd[1].fd = tunnel_in_fd (tunnel);
	  pollfd[1].events = POLLIN;

	  verbose ("poll ()");
	  n = poll (pollfd, 2, -1);
	  verbose (" = %d\n", n);
	  if (n == -1)
	    {
	      fprintf (stderr, "%s: poll error: %s\n",
		       arg.me, strerror (errno));
	      exit (1);
	    }
	  else if (n == 0)
	    {
	      debug ("poll() timed out\n");
	      continue;
	    }

	  handle_input ("device or port", tunnel, fd, pollfd[0].revents,
			handle_device_input, &closed);
	  handle_input ("tunnel", tunnel, fd, pollfd[1].revents,
			handle_tunnel_input, &closed);
#if 0
	  if (pollfd[0].revents)
	    {
	      ssize_t n;

	      n = handle_device_input (tunnel,
				       fd,
				       pollfd[0].revents);
	      if (n == 0 || (n == -1 && errno != EAGAIN))
		{
		  if (n == 0)
		    debug ("device or port closed\n");
		  else
		    debug ("device or port read error: %s\n",
			   strerror (errno));
		  closed = TRUE;
		}		
	    }

	  if (pollfd[1].revents)
	    {
	      ssize_t n;

	      n = handle_tunnel_input (tunnel,
				       pollfd[1].revents,
				       fd);
	      if (n == 0 || (n == -1 && errno != EAGAIN))
		{
		  if (n == 0)
		    debug ("tunnel closed by peer\n");
		  else
		    debug ("tunnel read error: %s\n",
			   strerror (errno));
		  closed = TRUE;
		}
	    }
#endif
	}

      debug ("closing tunnel\n");
      close (fd);
      tunnel_close (tunnel);
    }

  debug ("destroying tunnel\n");
  tunnel_destroy (tunnel);
 
  return 0;
}
